@@ -5,7 +5,7 @@ from array import array
 import numpy as np
 
 from traitlets import (
-    Bool, Unicode, List, Instance, CFloat, Tuple, Union, default
+    Bool, Dict, Enum, Unicode, List, Instance, CFloat, Tuple, TraitError, Union, default, validate
 )
 from traittypes import Array
 from ipywidgets import (
@@ -14,10 +14,12 @@ from ipywidgets import (
     Color, Image
 )
 
-from .serialization import array_serialization, component_array_serialization
+from .serialization import array_serialization, data_array_serialization
 
 from ._frontend import module_version, module_name
 
+from .colormaps import colormaps
+
 FLOAT32 = 'f'
 UINT32 = 'I'
 
@@ -42,12 +44,20 @@ class Component(_GanyWidgetBase):
     _model_name = Unicode('ComponentModel').tag(sync=True)
 
     name = Unicode().tag(sync=True)
-    array = Union((Instance(Widget), Array())).tag(sync=True, **component_array_serialization)
+    array = Union((Instance(Widget), Array())).tag(sync=True, **data_array_serialization)
 
     min = CFloat(allow_none=True, default_value=None)
     max = CFloat(allow_none=True, default_value=None)
 
-    # TODO Compute min and max in the constructor?
+    def __init__(self, name, array, **kwargs):
+        """Create a new Component instance given its name and array."""
+        super(Component, self).__init__(name=name, array=array, **kwargs)
+
+        if self.min is None:
+            self.min = np.min(self.array) if not isinstance(self.array, Widget) else np.min(self.array.array)
+
+        if self.max is None:
+            self.max = np.max(self.array) if not isinstance(self.array, Widget) else np.max(self.array.array)
 
 
 class Data(_GanyWidgetBase):
@@ -56,7 +66,20 @@ class Data(_GanyWidgetBase):
     _model_name = Unicode('DataModel').tag(sync=True)
 
     name = Unicode().tag(sync=True)
-    components = List(Instance(Component)).tag(sync=True, **widget_serialization)
+    components = Union((Dict(), List(Instance(Component)))).tag(sync=True, **widget_serialization)
+
+    def __init__(self, name, components, **kwargs):
+        """Create a new Data instance given its name and components."""
+        super(Data, self).__init__(name=name, components=components, **kwargs)
+
+    @property
+    def dim(self):
+        """Get the data dimension."""
+        return len(self.components)
+
+    def as_input(self):
+        """Internal method of ipygany. Do not use this."""
+        return [(self.name, comp.name) for comp in self.components]
 
     def __getitem__(self, key):
         """Get a component by name or index."""
@@ -71,18 +94,24 @@ class Data(_GanyWidgetBase):
 
         raise KeyError('Invalid key {}.'.format(key))
 
+    @validate('components')
+    def _validate_components(self, proposal):
+        components = proposal['value']
+
+        if isinstance(components, dict):
+            return [Component(name, array) for name, array in components.items()]
+
+        return components
+
 
 def _grid_data_to_data_widget(grid_data):
     """Turn a vtk grid into Data widgets."""
     data = []
     for key, value in grid_data.items():
-        d = Data(
-            name=key,
-            components=[
-                Component(name=comp_name, array=comp['array'])
-                for comp_name, comp in value.items()
-            ]
-        )
+        d = Data(key, [
+            Component(comp_name, comp['array'])
+            for comp_name, comp in value.items()
+        ])
         data.append(d)
 
     return data
@@ -104,33 +133,54 @@ class Block(_GanyWidgetBase):
 
     _model_name = Unicode('BlockModel').tag(sync=True)
 
-    vertices = Array(default_value=array(FLOAT32)).tag(sync=True, **array_serialization)
+    vertices = Union((Instance(Widget), Array()), default_value=array(FLOAT32)).tag(sync=True, **data_array_serialization)
 
     default_color = Color('#6395b0').tag(sync=True)
 
-    data = List(Instance(Data), default_value=[]).tag(sync=True, **widget_serialization)
+    data = Union((Dict(), List(Instance(Data)))).tag(sync=True, **widget_serialization)
 
     environment_meshes = List(Instance(Widget), default_value=[]).tag(sync=True, **widget_serialization)
 
     def __getitem__(self, key):
         """Get a component by name or index."""
-        if not isinstance(key, tuple) or len(key) != 2:
-            raise KeyError('You can only access data by (data_name, component_name) tuple.')
+        if not (isinstance(key, str) or (isinstance(key, tuple) and len(key) == 2)):
+            raise KeyError('You can only access data by (data_name, component_name) tuple or data_name string.')
 
+        # This prevents failures when this method is called in the constructor
+        # self.data is not yet initialized
+        actual_data = self.data if len(self.data) else (self.parent.data if self.parent else [])
+
+        # If the key is a string, we assume it's the data name
+        if isinstance(key, str):
+            for data in actual_data:
+                if data.name == key:
+                    return data
+            raise KeyError('Data {} not found.'.format(key))
+
+        # Otherwise it's a (data name, component name) tuple
         data_name = key[0]
         component_name = key[1]
 
         if isinstance(data_name, str):
-            for data in self.data:
+            for data in actual_data:
                 if data.name == data_name:
                     return data[component_name]
             raise KeyError('Data {} not found.'.format(data_name))
 
         if isinstance(data_name, int):
-            return self.data[data_name][component_name]
+            return actual_data[data_name][component_name]
 
         raise KeyError('Invalid key {}.'.format(key))
 
+    @validate('data')
+    def _validate_data(self, proposal):
+        data = proposal['value']
+
+        if isinstance(data, dict):
+            return [Data(name, components) for name, components in data.items()]
+
+        return data
+
 
 class PolyMesh(Block):
     """A polygon-based 3-D Mesh widget."""
@@ -145,12 +195,15 @@ class PolyMesh(Block):
         A PolyMesh is a triangle-based mesh. ``vertices`` is the array of points, ``triangle_indices`` is the array of triangle
         indices.
         """
-        vertices = np.asarray(vertices).flatten()
+        if not isinstance(vertices, Widget):
+            vertices = np.asarray(vertices).flatten()
         triangle_indices = np.asarray(triangle_indices).flatten()
 
         # If there are no triangle indices, assume vertices are given in the right order for constructing the triangles.
         if triangle_indices.size == 0:
-            triangle_indices = np.arange(vertices.size, dtype=np.uint32)
+            l_vertices = np.asarray(vertices.array).flatten().size if isinstance(vertices, Widget) else vertices.size
+
+            triangle_indices = np.arange(l_vertices, dtype=np.uint32)
 
         super(PolyMesh, self).__init__(
             vertices=vertices, triangle_indices=triangle_indices, data=data, **kwargs
@@ -364,26 +417,131 @@ class Effect(Block):
 
     _model_name = Unicode('EffectModel').tag(sync=True)
 
+    input = Union((Tuple(), Unicode(), CFloat())).tag(sync=True)
+
     parent = Instance(Block).tag(sync=True, **widget_serialization)
 
     def __init__(self, parent, **kwargs):
         """Create an Effect on the given Mesh or Effect output."""
         super(Effect, self).__init__(parent=parent, **kwargs)
 
+    @property
+    def data(self):
+        """Get data."""
+        return self.parent.data
+
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 0
+
+    @default('input')
+    def _default_input(self):
+        if not len(self.data):
+            if self.input_dim == 0 or self.input_dim == 1:
+                return 0
+            return tuple(0 for _ in range(self.input_dim))
+
+        return self._validate_input_impl(self.data[0].name)
+
+    @validate('input')
+    def _validate_input(self, proposal):
+        return self._validate_input_impl(proposal['value'])
+
+    def _validate_input_impl(self, value):
+        # Input is a data name
+        if isinstance(value, str):
+            input_data = self[value]
+
+            # Simply use this data
+            if input_data.dim == self.input_dim:
+                return input_data.name
+
+            # Take all the components and fill in with zeros
+            if input_data.dim < self.input_dim:
+                chosen_input = input_data.as_input()
+
+                while len(chosen_input) != self.input_dim:
+                    chosen_input.append(0.)
+
+                return chosen_input
+
+            # input_data.dim > self.input_dim, take only the first self.input_dim components
+            return input_data.as_input()[:self.input_dim]
+
+        # Input as a tuple
+        if isinstance(value, (tuple, list)):
+            if self.input_dim == 1 and len(value) == 2:
+                return self._validate_input_component(value)
+
+            if len(value) != self.input_dim:
+                raise TraitError('input is of dimension {} but expected input dimension is {}'.format(len(value), self.input_dim))
+
+            # Check all elements in the tuple
+            return tuple(self._validate_input_component(el) for el in value)
+
+        # Input is a number
+        if isinstance(value, (float, int)) and self.input_dim == 1:
+            return value
+
+        raise TraitError('{} is not a valid input'.format(value))
+
+    def _validate_input_component(self, value):
+        # Component selection by name
+        if isinstance(value, (tuple, list)):
+            if len(value) != 2:
+                raise TraitError('{} is not a valid component'.format(value))
+
+            try:
+                self[value[0], value[1]]
+            except KeyError:
+                raise TraitError('{} is not a valid component'.format(value))
+
+            return value
+
+        # Data selection by name
+        if isinstance(value, str):
+            try:
+                data = self[value]
+            except KeyError:
+                raise TraitError('{} is not a valid data'.format(value))
+
+            if data.dim != 1:
+                raise TraitError('{} is ambiguous, please select a component'.format(value))
+
+            return (data.name, data.components[0].name)
+
+        if isinstance(value, (float, int)):
+            return value
+
+        raise TraitError('{} is not a valid input'.format(value))
+
 
 class Warp(Effect):
     """A warp effect to another block."""
 
     _model_name = Unicode('WarpModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
-
     offset = Union((Tuple(trait=Unicode, minlen=3, maxlen=3), CFloat(0.)), default_value=0.).tag(sync=True)
     factor = Union((Tuple(trait=Unicode, minlen=3, maxlen=3), CFloat(0.)), default_value=1.).tag(sync=True)
 
-    @default('input')
-    def _default_input(self):
-        return self.parent.data[0].name
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 3
+
+
+class WarpByScalar(Effect):
+    """A warp-by-scalar effect to another block."""
+
+    _model_name = Unicode('WarpByScalarModel').tag(sync=True)
+
+    factor = CFloat(1.).tag(sync=True)
+
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
 
 
 class Alpha(Effect):
@@ -391,19 +549,25 @@ class Alpha(Effect):
 
     _model_name = Unicode('AlphaModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
-
     @default('input')
     def _default_input(self):
         return 0.7
 
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
+
 
 class RGB(Effect):
     """A color effect to another block."""
 
     _model_name = Unicode('RGBModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 3
 
 
 class IsoColor(Effect):
@@ -411,14 +575,32 @@ class IsoColor(Effect):
 
     _model_name = Unicode('IsoColorModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
-
     min = CFloat(0.).tag(sync=True)
     max = CFloat(0.).tag(sync=True)
+    range = Tuple((0., 0.)).tag(sync=True)
+    colormap = Enum(list(colormaps.values()), allow_none=False, default_value=colormaps.Viridis).tag(sync=True)
+    type = Enum(['linear', 'log'], default_value='linear').tag(sync=True)
 
-    @default('input')
-    def _default_input(self):
-        return self.parent.data[0].name
+    def __init__(self, parent, **kwargs):
+        super().__init__(parent, **kwargs)
+        self.range = (self.min, self.max)
+
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
+
+
+class ColorBar(_GanyDOMWidgetBase):
+    """A ColorBar widget."""
+
+    _view_name = Unicode('ColorBarView').tag(sync=True)
+    _model_name = Unicode('ColorBarModel').tag(sync=True)
+
+    parent = Instance(IsoColor, allow_none=False).tag(sync=True, **widget_serialization)
+
+    def __init__(self, parent, **kwargs):
+        super(ColorBar, self).__init__(parent=parent, **kwargs)
 
 
 class IsoSurface(Effect):
@@ -426,14 +608,13 @@ class IsoSurface(Effect):
 
     _model_name = Unicode('IsoSurfaceModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
-
     value = CFloat(0.).tag(sync=True)
     dynamic = Bool(False).tag(sync=True)
 
-    @default('input')
-    def _default_input(self):
-        return self.parent.data[0].name
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
 
 
 class Threshold(Effect):
@@ -441,15 +622,20 @@ class Threshold(Effect):
 
     _model_name = Unicode('ThresholdModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode(), CFloat(0.))).tag(sync=True)
-
     min = CFloat(0.).tag(sync=True)
     max = CFloat(0.).tag(sync=True)
+    range = Tuple((0., 0.)).tag(sync=True)
     dynamic = Bool(False).tag(sync=True)
+    inclusive = Bool(True).tag(sync=True)
 
-    @default('input')
-    def _default_input(self):
-        return self.parent.data[0].name
+    def __init__(self, parent, **kwargs):
+        super().__init__(parent, **kwargs)
+        self.range = (self.min, self.max)
+
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
 
 
 class UnderWater(Effect):
@@ -457,14 +643,15 @@ class UnderWater(Effect):
 
     _model_name = Unicode('UnderWaterModel').tag(sync=True)
 
-    input = Union((Tuple(trait=Unicode, minlen=2, maxlen=2), Unicode())).tag(sync=True)
-
     default_color = Color('#F2FFD2').tag(sync=True)
     texture = Instance(Image, allow_none=True, default_value=None).tag(sync=True, **widget_serialization)
+    texture_scale = CFloat(2.).tag(sync=True)
+    texture_position = Tuple(minlen=2, maxlen=2, default_value=(1., 1., 0.)).tag(sync=True)
 
-    @default('input')
-    def _default_input(self):
-        return self.parent.data[0].name
+    @property
+    def input_dim(self):
+        """Input dimension."""
+        return 1
 
 
 class Water(Effect):
@@ -486,8 +673,10 @@ class Scene(_GanyDOMWidgetBase):
 
     children = List(Instance(Block)).tag(sync=True, **widget_serialization)
 
-    background_color = Color('#fff').tag(sync=True)
-    background_opacity = CFloat(0.).tag(sync=True)
+    background_color = Color('white').tag(sync=True)
+    background_opacity = CFloat(1.).tag(sync=True)
+
+    camera = Dict(allow_none=True, default_value=None).tag(sync=True)
 
     def __init__(self, children=[], **kwargs):
         """Construct a Scene."""
