@@ -4,16 +4,12 @@
 import * as THREE from 'three';
 
 import {
-  WidgetModel, DOMWidgetModel, DOMWidgetView, ISerializers, unpack_models
+  WidgetModel, DOMWidgetModel, DOMWidgetView, ISerializers, unpack_models, WidgetView, Dict
 } from '@jupyter-widgets/base';
 
-import {
-  ImageView
-} from '@jupyter-widgets/controls';
-
 import {
   Message
-} from '@phosphor/messaging';
+} from '@lumino/messaging';
 
 import {
   MODULE_NAME, MODULE_VERSION
@@ -27,10 +23,18 @@ import {
   Data, Component,
   Block, Effect,
   PolyMesh, TetraMesh, PointCloud,
-  Warp, Alpha, RGB, IsoColor, IsoSurface, Threshold,
+  Warp, WarpByScalar, Alpha, RGB, IsoColor, IsoSurface, Threshold,
   Water, UnderWater,
 } from 'ganyjs';
 
+const colormaps: string[] = [
+  'BrBG', 'PRGn', 'PiYG', 'PuOr', 'RdBu', 'RdGy', 'RdYlBu', 'RdYlGn',
+  'Spectral', 'BuGn', 'BuPu', 'GnBu', 'OrRd', 'PuBuGn', 'PuBu', 'PuRd', 'RdPu',
+  'YlGnBu', 'YlGn', 'YlOrBr', 'YlOrRd', 'Blues', 'Greens', 'Greys', 'Purples', 'Reds',
+  'Oranges',  'Cividis',  'CubehelixDefault',  'Rainbow',  'Warm',  'Cool',  'Sinebow',  'Turbo',
+  'Viridis', 'Magma', 'Inferno', 'Plasma'
+]
+
 
 function deserialize_float32array (data: any, manager: any) {
     return new Float32Array(data.data.buffer);
@@ -40,7 +44,7 @@ function deserialize_uint32array (data: any, manager: any) {
     return new Uint32Array(data.data.buffer);
 }
 
-function deserialize_component_array (value: any, manager: any) {
+function deserialize_data_array (value: any, manager: any) {
   if (typeof value == 'string') {
     return unpack_models(value, manager);
   } else {
@@ -98,7 +102,7 @@ class ComponentModel extends _GanyWidgetModel {
     const array = this.get('array');
 
     if (array.hasOwnProperty('name') && array.name == 'NDArrayModel') {
-      return array.getNDArray().data;
+      return new Float32Array(array.getNDArray().data);
     } else {
       return array;
     }
@@ -116,7 +120,7 @@ class ComponentModel extends _GanyWidgetModel {
 
   static serializers: ISerializers = {
     ..._GanyWidgetModel.serializers,
-    array: { deserialize: deserialize_component_array },
+    array: { deserialize: deserialize_data_array },
   }
 
   static model_name = 'ComponentModel';
@@ -178,8 +182,14 @@ abstract class BlockModel extends _GanyWidgetModel {
     this.initEventListeners();
   }
 
-  get vertices () : Float32Array {
-    return this.get('vertices');
+  get vertices () {
+    const array = this.get('vertices');
+
+    if (array.hasOwnProperty('name') && array.name == 'NDArrayModel') {
+      return new Float32Array(array.getNDArray().data);
+    } else {
+      return array;
+    }
   }
 
   get data () : Data[] {
@@ -204,7 +214,7 @@ abstract class BlockModel extends _GanyWidgetModel {
 
   static serializers: ISerializers = {
     ..._GanyWidgetModel.serializers,
-    vertices: { deserialize: deserialize_float32array },
+    vertices: { deserialize: deserialize_data_array },
     data: { deserialize: (unpack_models as any) },
     environment_meshes: { deserialize: (unpack_models as any) },
   }
@@ -403,6 +413,40 @@ class WarpModel extends EffectModel {
 }
 
 
+export
+class WarpByScalarModel extends EffectModel {
+
+  defaults() {
+    return {...super.defaults(),
+      _model_name: WarpByScalarModel.model_name,
+    };
+  }
+
+  get input () {
+    return this.get('input');
+  }
+
+  get factor () {
+    return this.get('factor');
+  }
+
+  createBlock () {
+    return new WarpByScalar(this.parent.block, this.input, this.factor);
+  }
+
+  initEventListeners () : void {
+    super.initEventListeners();
+
+    this.on('change:factor', () => { this.block.factor = this.factor; });
+  }
+
+  block: WarpByScalar;
+
+  static model_name = 'WarpByScalarModel';
+
+}
+
+
 export
 class AlphaModel extends EffectModel {
 
@@ -461,6 +505,9 @@ class IsoColorModel extends EffectModel {
       _model_name: IsoColorModel.model_name,
       min: 0.,
       max: 0.,
+      range: [0., 0.],
+      colormap: 34,
+      type: 'linear',
     };
   }
 
@@ -468,10 +515,34 @@ class IsoColorModel extends EffectModel {
     return this.get('min');
   }
 
+  set min (value: number) {
+    this.set('min', value)
+  }
+
   get max () {
     return this.get('max');
   }
 
+  set max (value: number) {
+    this.set('max', value)
+  }
+
+  get range () {
+    return this.get('range');
+  }
+
+  set range (value: [number, number]) {
+    this.set('range', value)
+  }
+
+  get colormap () : string {
+    return colormaps[this.get('colormap')];
+  }
+
+  get type () {
+    return this.get('type');
+  }
+
   get input () {
     const input = this.get('input');
 
@@ -479,14 +550,26 @@ class IsoColorModel extends EffectModel {
   }
 
   createBlock () {
-    return new IsoColor(this.parent.block, this.input, this.min, this.max);
+    return new IsoColor(this.parent.block, this.input, this.min, this.max, this.colormap, this.type);
   }
 
   initEventListeners () : void {
     super.initEventListeners();
 
-    this.on('change:min', () => { this.block.min = this.min; });
-    this.on('change:max', () => { this.block.max = this.max; });
+    this.on('change:min', () => {
+      this.block.min = this.min;
+      this.range = [this.min, this.range[1]];
+    });
+    this.on('change:max', () => {
+      this.block.max = this.max;
+      this.range = [this.range[0], this.max];
+    });
+    this.on('change:range', () => {
+      this.min = this.range[0];
+      this.max = this.range[1];
+    });
+    this.on('change:colormap', () => { this.block.colorMap = this.colormap; });
+    this.on('change:type', () => { this.block.type = this.type; });
   }
 
   block: IsoColor;
@@ -496,6 +579,96 @@ class IsoColorModel extends EffectModel {
 }
 
 
+export
+class ColorBarModel extends _GanyDOMWidgetModel {
+
+  defaults() {
+    return {...super.defaults(),
+      _model_name: ColorBarModel.model_name,
+      _view_name: ColorBarModel.view_name,
+      parent: null,
+    };
+  }
+
+  initialize (attributes: any, options: any) {
+    super.initialize(attributes, options);
+
+    this.colorBar = this.parent.block.colorBar;
+    this.parent.block.on('change:colorbar', this.redraw.bind(this));
+    this.redraw();
+
+    this.on('change:parent', this.updateColorBar.bind(this));
+  }
+
+  get parent () : IsoColorModel {
+    return this.get('parent')
+  }
+
+  private updateColorBar () {
+    // TODO Turn off events on old parent
+    this.colorBar = this.parent.block.colorBar;
+    this.parent.block.on('change:colorbar', this.redraw.bind(this));
+    this.redraw();
+  }
+
+  private redraw () {
+    for (const view_id in this.views) {
+      this.views[view_id].then((view: ColorBarView) => {
+        view.draw();
+      });
+    }
+  }
+
+  static serializers: ISerializers = {
+    ..._GanyDOMWidgetModel.serializers,
+    parent: { deserialize: (unpack_models as any) },
+  }
+
+  static model_name = 'ColorBarModel';
+  static view_name = 'ColorBarView';
+
+  colorBar: HTMLCanvasElement;
+
+  views: Dict<Promise<ColorBarView>>;
+
+}
+
+
+export
+class ColorBarView extends DOMWidgetView {
+
+  render () {
+    const ctx = this.el.getContext('2d');
+    this.el.classList.add('gany-colorbar');
+
+    if (ctx === null) {
+      throw 'Failed to create canvas context for ColorBarView';
+    }
+
+    this.ctx = ctx;
+
+    this.draw();
+  }
+
+  draw () {
+    this.el.width = this.model.colorBar.width;
+    this.el.height = this.model.colorBar.height;
+    this.ctx.drawImage(this.model.colorBar, 0, 0);
+  }
+
+  // @ts-ignore
+  get tagName(): string {
+    return 'canvas';
+  }
+
+  el: HTMLCanvasElement;
+  ctx: CanvasRenderingContext2D;
+
+  model: ColorBarModel;
+
+}
+
+
 export
 class IsoSurfaceModel extends EffectModel {
 
@@ -546,7 +719,9 @@ class ThresholdModel extends EffectModel {
       _model_name: ThresholdModel.model_name,
       min: 0.,
       max: 0.,
+      range: [0., 0.],
       dynamic: false,
+      inclusive: true,
     };
   }
 
@@ -554,14 +729,34 @@ class ThresholdModel extends EffectModel {
     return this.get('min');
   }
 
+  set min (value: number) {
+    this.set('min', value)
+  }
+
   get max () {
     return this.get('max');
   }
 
+  set max (value: number) {
+    this.set('max', value)
+  }
+
+  get range () {
+    return this.get('range');
+  }
+
+  set range (value: [number, number]) {
+    this.set('range', value)
+  }
+
   get dynamic () {
     return this.get('dynamic');
   }
 
+  get inclusive (): boolean {
+    return this.get('inclusive');
+  }
+
   get input () {
     const input = this.get('input');
 
@@ -569,14 +764,28 @@ class ThresholdModel extends EffectModel {
   }
 
   createBlock () {
-    return new Threshold(this.parent.block, this.input, {min: this.min, max: this.max, dynamic: this.dynamic});
+    return new Threshold(this.parent.block, this.input, {
+      min: this.min, max: this.max,
+      dynamic: this.dynamic, inclusive: this.inclusive
+    });
   }
 
   initEventListeners () : void {
     super.initEventListeners();
 
-    this.on('change:min', () => { this.block.min = this.min });
-    this.on('change:max', () => { this.block.max = this.max });
+    this.on('change:min', () => {
+      this.block.min = this.min;
+      this.range = [this.min, this.range[1]];
+    });
+    this.on('change:max', () => {
+      this.block.max = this.max;
+      this.range = [this.range[0], this.max];
+    });
+    this.on('change:range', () => {
+      this.min = this.range[0];
+      this.max = this.range[1];
+    });
+    this.on('change:inclusive', () => { this.block.inclusive = this.inclusive });
   }
 
   block: Threshold;
@@ -594,6 +803,8 @@ class UnderWaterModel extends EffectModel {
       _model_name: UnderWaterModel.model_name,
       default_color: '#F2FFD2',
       texture: null,
+      texture_scale: 2.,
+      texture_position: [1., 1., 0.],
     };
   }
 
@@ -609,8 +820,13 @@ class UnderWaterModel extends EffectModel {
     return typeof input == 'string' ? input : [input];
   }
 
-  get defaultcolor () {
-    return new THREE.Color(this.get('default_color'));
+  get textureScale () {
+    return this.get('texture_scale');
+  }
+
+  get texturePosition (): THREE.Vector3 {
+    const texturePosition = this.get('texture_position');
+    return new THREE.Vector3(texturePosition[0], texturePosition[1], texturePosition[2]);
   }
 
   setTexture () {
@@ -622,7 +838,7 @@ class UnderWaterModel extends EffectModel {
       return;
     }
 
-    this.widget_manager.create_view(image, {}).then((imageView: ImageView) => {
+    this.widget_manager.create_view(image, {}).then((imageView: WidgetView) => {
       const textureLoader = new THREE.TextureLoader();
 
       textureLoader.load(imageView.el.src, (texture: THREE.Texture) => {
@@ -632,7 +848,7 @@ class UnderWaterModel extends EffectModel {
   }
 
   createBlock () {
-    return new UnderWater(this.parent.block, this.input, { defaultColor: this.defaultColor });
+    return new UnderWater(this.parent.block, this.input, { defaultColor: this.defaultColor, textureScale: this.textureScale, texturePosition: this.texturePosition });
   }
 
   initEventListeners() : void {
@@ -641,6 +857,12 @@ class UnderWaterModel extends EffectModel {
     this.on('change:texture', () => {
       this.setTexture();
     });
+    this.on('change:texture_scale', () => {
+      this.block.textureScale = this.textureScale;
+    });
+    this.on('change:texture_position', () => {
+      this.block.texturePosition = this.texturePosition;
+    });
   }
 
   block: UnderWater;
@@ -680,10 +902,18 @@ class WaterModel extends EffectModel {
   }
 
   createBlock () {
+    // TODO Add texture to the package
+    const url = 'https://raw.githubusercontent.com/martinRenou/threejs-caustics/master/assets/TropicalSunnyDay_';
+
     return new Water(this.parent.block, {
       underWaterBlocks: this.underWaterBlocks,
       causticsEnabled: this.causticsEnabled,
       causticsFactor: this.causticsFactor,
+      skybox: new THREE.CubeTextureLoader().load([
+        url + 'px.jpg', url + 'nx.jpg',
+        url + 'py.jpg', url + 'ny.jpg',
+        url + 'pz.jpg', url + 'nz.jpg',
+      ]),
     });
   }
 
@@ -713,9 +943,10 @@ class SceneModel extends _GanyDOMWidgetModel {
     return {...super.defaults(),
       _model_name: SceneModel.model_name,
       _view_name: SceneModel.view_name,
-      background_color: '#fff',
-      background_opacity: 0.,
+      background_color: 'white',
+      background_opacity: 1.,
       children: [],
+      camera: null,
     };
   }
 
@@ -726,6 +957,7 @@ class SceneModel extends _GanyDOMWidgetModel {
 
     this.updateChildren();
     this.on('change:children', this.updateChildren.bind(this));
+    this.on('change:camera', this.updateCamera.bind(this));
   }
 
   get backgroundColor () : string {
@@ -736,6 +968,56 @@ class SceneModel extends _GanyDOMWidgetModel {
     return this.get('background_opacity')
   }
 
+  get camera () : any {
+    return this.get('camera');
+  }
+
+  set camera (value: any) {
+    this.set('camera', value);
+  }
+
+  updateCamera () {
+    const blocks: Block[] = this.get('children').map((child: BlockModel) => child.block);
+
+    if (blocks.length == 0) {
+      return;
+    }
+
+    // TODO Compute mean radius and mean center of all the blocks
+    const radius = blocks[0].boundingSphere.radius;
+    const center = blocks[0].boundingSphere.center;
+
+    const modelCameraPosition = this.camera ? this.camera['position'] || null : null;
+    const modelCameraRotation = this.camera ? this.camera['rotation'] || null : null;
+    const modelCameraTarget = this.camera ? this.camera['target'] || null : null;
+    const modelCameraUp = this.camera ? this.camera['up'] || null : null;
+
+    if (modelCameraPosition) {
+      this.cameraPosition = new THREE.Vector3(modelCameraPosition[0], modelCameraPosition[1], modelCameraPosition[2]);
+    } else {
+      this.cameraPosition = new THREE.Vector3(center.x, center.y, center.z + 2.5 * radius);
+    }
+
+    if (modelCameraRotation) {
+      this.cameraRotation = new THREE.Quaternion(modelCameraRotation[0], modelCameraRotation[1], modelCameraRotation[2], modelCameraRotation[3]);
+    }
+
+    if (modelCameraTarget) {
+      this.cameraTarget = new THREE.Vector3(modelCameraTarget[0], modelCameraTarget[1], modelCameraTarget[2]);
+    } else {
+      this.cameraTarget = new THREE.Vector3(center.x, center.y, center.z);
+    }
+
+    if (modelCameraUp) {
+      this.cameraUp = new THREE.Vector3(modelCameraUp[0], modelCameraUp[1], modelCameraUp[2]);
+    }
+
+    this.cameraNear = radius / 3.;
+    this.cameraFar = radius * 10.;
+
+    this.trigger('update_camera');
+  }
+
   private updateChildren () {
     // TODO: Remove old children
 
@@ -745,21 +1027,22 @@ class SceneModel extends _GanyDOMWidgetModel {
       return;
     }
 
-    const boundingSphereRadius = Math.max(...blocks.map((block: Block) => block.boundingSphere.radius));
-    const scale = new THREE.Vector3(1 / boundingSphereRadius, 1 / boundingSphereRadius, 1 / boundingSphereRadius);
-
-    const position = blocks[0].boundingSphere.center;
-
     for (const block of blocks) {
-      block.scale = scale;
-      block.position = new THREE.Vector3(-position.x, -position.y, -position.z);
-
       this.scene.addBlock(block);
     }
+
+    this.updateCamera();
   }
 
   scene: Scene;
 
+  cameraPosition: THREE.Vector3;
+  cameraRotation: THREE.Quaternion;
+  cameraTarget: THREE.Vector3;
+  cameraUp: THREE.Vector3;
+  cameraNear: number = 0.01;
+  cameraFar: number = 99999;
+
   static serializers: ISerializers = {
     ..._GanyDOMWidgetModel.serializers,
     children: { deserialize: (unpack_models as any) },
@@ -784,6 +1067,8 @@ class SceneView extends DOMWidgetView {
       this.renderer.backgroundColor = this.model.backgroundColor
       this.renderer.backgroundOpacity = this.model.backgroundOpacity;
 
+      this.updateCamera();
+
       this.initEventListeners();
     });
   }
@@ -793,6 +1078,36 @@ class SceneView extends DOMWidgetView {
 
     this.model.on('change:background_color', () => { this.renderer.backgroundColor = this.model.backgroundColor; });
     this.model.on('change:background_opacity', () => { this.renderer.backgroundOpacity = this.model.backgroundOpacity; });
+
+    this.model.on('update_camera', this.updateCamera.bind(this));
+
+    this.renderer.controls.addEventListener('change', this.handleCameraMove.bind(this));
+  }
+
+  updateCamera () {
+    if (this.model.cameraPosition && !this.model.cameraPosition.equals(this.renderer.cameraPosition)) {
+      this.renderer.cameraPosition = this.model.cameraPosition;
+    }
+    if (this.model.cameraRotation && !this.model.cameraRotation.equals(this.renderer.cameraRotation)) {
+      this.renderer.cameraRotation = this.model.cameraRotation;
+    }
+    if (this.model.cameraTarget && !this.model.cameraTarget.equals(this.renderer.cameraTarget)) {
+      this.renderer.cameraTarget = this.model.cameraTarget;
+    }
+    if (this.model.cameraUp && !this.model.cameraUp.equals(this.renderer.cameraUp)) {
+      this.renderer.cameraUp = this.model.cameraUp;
+    }
+    this.renderer.camera.near = this.model.cameraNear;
+    this.renderer.camera.far = this.model.cameraFar;
+  }
+
+  handleCameraMove() {
+    this.model.camera = {
+      position: this.renderer.cameraPosition.toArray(),
+      rotation: this.renderer.cameraRotation.toArray(),
+      target: this.renderer.cameraTarget.toArray(),
+      up: this.renderer.cameraUp.toArray(),
+    };
   }
 
   processPhosphorMessage (msg: Message) {
